import re

from lms.message.definitions.field.lms_field import LMS_Field
from lms.message.tag.lms_tag import (LMS_ControlTag, LMS_DecodedTag,
                                     LMS_EncodedTag)
from lms.titleconfig.config import TagConfig


class LMS_MessageText:
    """Class that represents a message text entry."""

    TAG_FORMAT = re.compile(r"(\[[^\]]+\])")

    def __init__(
        self, message: str | list[str | LMS_ControlTag], config: TagConfig | None = None
    ):
        if isinstance(message, str):
            self._set_parts(message)
        else:
            self._parts = message

        self._config = config

    def __iter__(self):
        return iter(self._parts)

    @property
    def text(self) -> str:
        """The raw text of the message."""
        text_list = []
        for part in self._parts:
            text_list.append(
                part.to_text()
                if isinstance(part, (LMS_EncodedTag, LMS_DecodedTag))
                else part
            )
        return "".join(text_list)

    @text.setter
    def text(self, string: str):
        self._set_parts(string)

    @property
    def tags(self) -> list[LMS_ControlTag]:
        """The list of control tags in the message."""
        return [
            part
            for part in self._parts
            if isinstance(part, (LMS_EncodedTag, LMS_DecodedTag))
        ]

    def append_encoded_tag(self, group_id: int, tag_index: int, *parameters: str):
        """Appends an encoded tag to the current message.

        :param group: the group name or index.
        :param tag: the group tag or index:
        :param parameters: a list of hex strings.

        ## Usage
        ```
        message.append_encoded_tag(1, 2, "01", "00", "00", "CD")
        ```
        """
        self._parts.append(LMS_EncodedTag(group_id, tag_index, list(parameters)))

    def append_decoded_tag(
        self,
        group_name: str,
        tag_name: str,
        **parameters: int | str | float | bool | bytes,
    ) -> None:
        """Appends an decoded tag to the current message.

        :param group_name: the group name.
        :param tag_name: the tag name.:
        :param parameters: keyword arguments of parameters mapped to their value.

        ## Usage
        ```
        message.append_decoded_tag("Mii", "Nickname", buffer=1, type="Voice", conversion="None")
        ```
        """
        if self._config is None:
            raise ValueError("A TitleConfig is required to append decoded tags.")

        definition = self._config.get_definition_by_names(group_name, tag_name)

        converted_params = {}
        for param_def in definition.parameters:
            converted_params[param_def.name] = LMS_Field(
                parameters[param_def.name], param_def
            )

        self._parts.append(LMS_DecodedTag(definition, converted_params))

    def append_tag_string(self, tag: str) -> None:
        """Appends a tag to the current message given a string.

        :param tag: the tag string."""
        if re.match(LMS_DecodedTag.TAG_FORMAT, tag):
            if self._config is None:
                raise ValueError("TagConfig is required to append decoded tags.")
            self._parts.append(LMS_DecodedTag.from_string(tag, self._config))
        elif re.match(LMS_EncodedTag.TAG_FORMAT, tag):
            self._parts.append(LMS_EncodedTag.from_string(tag))
        else:
            raise ValueError(f"Invalid format for tag '{tag}'.")

    def _set_parts(self, text: str) -> None:
        self._parts = []
        for part in self.TAG_FORMAT.split(text):
            if bool(re.match(self.TAG_FORMAT, part)):
                self.append_tag_string(part)
            else:
                self._parts.append(part)



from typing import Iterable, cast

from lms.common.lms_datatype import LMS_DataType
from lms.titleconfig.definitions.value import ValueDefinition

# Typehint that represents a map of strings to field instances.
# Utilized as a means of storing attributes, or parameters in decoded tags mapped to their string names.
type LMS_FieldMap = dict[str, LMS_Field]


def dict_to_field_map(data: dict, definitions: Iterable[ValueDefinition]):
    return {
        definition.name: LMS_Field(data[definition.name], definition)
        for definition in definitions
    }


class LMS_Field:
    """A class that represents a mapped field linked to a config definition.

    Acts as values for Attributes and Tag Parameters."""

    def __init__(
        self, value: int | str | float | bytes | bool, definition: ValueDefinition
    ):
        self._definition = definition
        self._verify_value(value)
        self._value = value

    def __repr__(self):
        if self.datatype is LMS_DataType.LIST:
            if len(self.list_items) > 6:
                preview = self.list_items[:3] + ["..."]
            else:
                preview = self.list_items
            return f"LMS_Field(value={self._value}, options={preview})"
        return f"LMS_Field(value={self._value}, type={self.datatype.name})"

    @property
    def name(self) -> str:
        """The name of the field."""
        return self._definition.name

    @property
    def description(self) -> str:
        """The description of the field."""
        return self._definition.description

    @property
    def value(self) -> int | str | float | bytes | bool:
        """The value of the field instance."""
        return self._value

    @property
    def datatype(self) -> LMS_DataType:
        """The datatype of the field instance."""
        return self._definition.datatype

    @property
    def list_items(self) -> list[str]:
        """The list items bound to the field instance. Only is valid for LMS_Datatype.LIST values."""
        return self._definition.list_items

    @value.setter
    def value(self, new_value: int | str | float | bytes | bool):
        self._verify_value(new_value)
        self._value = new_value

    def _verify_value(self, value: int | str | float | bytes | bool) -> None:
        datatype = self._definition.datatype

        if datatype in (LMS_DataType.BOOL, LMS_DataType.STRING):
            return

        match datatype:
            case LMS_DataType.BYTES if isinstance(value, bytes):
                if len(value) != 1:
                    raise ValueError("Byte types only work for values of length 1!")
                else:
                    return
            case LMS_DataType.LIST if isinstance(value, str):
                if value not in self._definition.list_items:
                    raise ValueError(
                        f"The value of '{value}' provided for field '{self.name}' is not in the list {self._definition.list_items}."
                    )
                else:
                    return
            case LMS_DataType.FLOAT32 if isinstance(value, float):
                _verify_range(value, -3.4028235e38, 3.4028235e38, self._definition)
                return
            case _ if isinstance(value, int):
                bits = datatype.stream_size * 8
                if datatype.signed:
                    max_value = 2 ** (bits - 1)
                    min_value = -max_value
                else:
                    min_value, max_value = 0, (2**bits) - 1

                _verify_range(value, min_value, max_value, self._definition)
                return

        raise TypeError(
            f"The value provided for '{self.name}' type '{type(value)}' should be '{self._definition.datatype.to_string()}'."
        )


def convert_string_to_type(
    value: str, datatype: LMS_DataType
) -> int | str | bool | float | bytes:
    if datatype in (LMS_DataType.STRING, LMS_DataType.LIST):
        return cast(str, value)

    match datatype:
        case LMS_DataType.BYTES:
            return bytes.fromhex(value)
        case LMS_DataType.BOOL:
            if value not in ("false", "true"):
                raise ValueError("Value must be true or false for bool type.")
            return value.strip().lower() == "true"
        case LMS_DataType.FLOAT32:
            return float(value)
        case _:
            return int(value)


def _verify_range(
    value: int | float, min: int | float, max: int | float, definition: ValueDefinition
):
    if not min <= value <= max:
        raise ValueError(
            f"The value '{value}' of type '{definition.datatype}' provided for field '{definition.name}' is out of range of ({min}, {max})"
        )



from dataclasses import dataclass

from lms.fileio.encoding import FileEncoding


@dataclass(frozen=True)
class LMS_FileInfo:
    is_big_endian: bool
    encoding: FileEncoding
    version: int
    section_count: int



from typing import Any, Self

from lms.message.definitions.field.lms_field import (LMS_Field, LMS_FieldMap,
                                                     dict_to_field_map)
from lms.message.definitions.lms_messagetext import LMS_MessageText
from lms.titleconfig.definitions.attribute import AttributeConfig
from lms.titleconfig.definitions.tags import TagConfig


class MSBTEntry:
    """A class that represents an entry in a MSBT file."""

    def __init__(
        self,
        name: str,
        message: str | LMS_MessageText,
        attribute: LMS_FieldMap | bytes | None = None,
        style_index: int | None = None,
    ):
        self.name = name

        if isinstance(message, str):
            self._message = LMS_MessageText(message)
        else:
            self._message = message

        self._attribute = attribute
        self.style_index = style_index

    @property
    def message(self) -> LMS_MessageText:
        """The message object for the instance."""
        return self._message

    @property
    def attribute(self) -> LMS_FieldMap | bytes | None:
        """The attribute for the instance."""
        return self._attribute

    def to_dict(self) -> dict:
        """Converts the MSBTEntry instance into a dictionary object."""
        result = {}
        result["name"] = self.name
        result["message"] = "" if self.message is None else self.message.text

        if self.attribute is not None:
            if isinstance(self.attribute, bytes):
                result["attribute"] = self.attribute.hex().upper()
            else:
                result["attribute"] = {name: attr.value for name, attr in self.attribute.items()}  # type: ignore

        result["style_index"] = self.style_index
        return result

    @classmethod
    def from_dict(
        cls,
        data: dict,
        attribute_conifg: AttributeConfig | None = None,
        tag_config: TagConfig | None = None,
    ) -> Self:
        """Creates a MSBTEntry from a dictionary object.

        :param data: the dictionary data.
        :param attribute_config: the config to use to import decoded attributes.
        :param tag_config: the config to use if decoded tags are included in the message.
        """
        if "message" in data:
            message = LMS_MessageText(data["message"], tag_config)

        attribute = None
        if "attribute" in data:
            attr = data["attribute"]
            if isinstance(attr, dict):
                if attribute_conifg is None:
                    raise TypeError(
                        "A valid attribute config must be provided for decoded attributes!"
                    )
                attribute = dict_to_field_map(attr, attribute_conifg.definitions)
            elif isinstance(attr, str):
                attribute = bytes.fromhex(attr)
            else:
                raise TypeError("Invalid attribute type!")

        style_index = data.get("style_index")
        return cls(data["name"], message, attribute, style_index)



from dataclasses import dataclass, field

from lms.titleconfig.definitions.value import ValueDefinition


@dataclass(frozen=True)
class AttributeConfig:
    """Class that represents an attribute config definition"""

    name: str
    description: str
    definitions: list[ValueDefinition]



from lms.fileio.io import FileReader
from lms.project.definitions.style import LMS_Style


def read_styles(reader: FileReader) -> list[LMS_Style]:
    style_list = []

    count = reader.read_uint32()
    for _ in range(count):
        region_width = reader.read_uint32()
        line_number = reader.read_uint32()
        font_index = reader.read_uint32()
        color_index = reader.read_uint32()

        style_list.append(LMS_Style(region_width, line_number, font_index, color_index))

    return style_list



from lms.fileio.io import FileReader
from lms.project.definitions.color import LMS_Color


def read_clr1(reader: FileReader) -> list[LMS_Color]:
    count = reader.read_uint32()
    return [
        LMS_Color(
            reader.read_uint32(),
            reader.read_uint32(),
            reader.read_uint32(),
            reader.read_uint32(),
        )
        for _ in range(count)
    ]



from typing import BinaryIO

from lms.common import lms_exceptions
from lms.common.stream.fileinfo import read_file_info, write_file_info
from lms.common.stream.hashtable import read_labels, write_labels
from lms.common.stream.section import (read_section_data, write_section,
                                       write_unsupported_section)
from lms.fileio.io import FileReader, FileWriter
from lms.message.msbt import MSBT
from lms.message.msbtentry import MSBTEntry
from lms.message.section.atr1 import (read_decoded_atr1, read_encoded_atr1,
                                      write_decoded_atr1, write_encoded_atr1)
from lms.message.section.tsy1 import read_tsy1, write_tsy1
from lms.message.section.txt2 import read_txt2, write_txt2
from lms.titleconfig.config import AttributeConfig, TagConfig


def read_msbt_path(
    file_path: str,
    attribute_config: AttributeConfig | None = None,
    tag_config: TagConfig | None = None,
) -> MSBT:
    """
    Reads a MSBT file from a given path.

    :param file_path: the path to the MSBT file.
    :param attribute_config: the attribute config to use for decoding attributes.
    :param tag_config: the tag config to use for decoding tags.

    ## Usage
    ```
    msbt = read_msbt_path("path/to/file.msbt")
    ...
    ```
    """
    with open(file_path, "rb") as stream:
        return read_msbt(stream, attribute_config, tag_config)


def write_msbt_path(file_path: str, file: MSBT) -> None:
    """
    Writes a MSBT file to a given path.

    :param file_path: the path to the MSBT file.
    :param file: the MSBT file object.

    ## Usage
    ```
    write_msbt_path(file_path, msbt)
    ...
    ```
    """
    with open(file_path, "wb") as stream:
        write_msbt(stream, file)


def read_msbt(
    stream: BinaryIO,
    attribute_config: AttributeConfig | None = None,
    tag_config: TagConfig | None = None,
) -> MSBT:
    """
    Reads a MSBT file from a stream.

    :param stream: a stream object.
    :param attribute_config: the attribute config to use for decoding attributes.
    :param tag_config: the tag config to use for decoding tags.

    ## Usage
    ```
    with open(file_path, "rb") as file:
        msbt = read_msbt(file)
        ...
    ```
    """
    if stream is None:
        raise ValueError("Stream must be valid!")

    reader = FileReader(stream)
    file_info = read_file_info(reader, "MsgStdBn")

    file = MSBT(file_info, attribute_config, tag_config)

    if attribute_config is not None:
        file.encoded_attributes = False

    attributes = style_indexes = None
    for magic, size in read_section_data(reader, file_info.section_count):
        match magic:
            case "LBL1":
                labels, slot_count = read_labels(reader)
                file.slot_count = slot_count
            case "ATR1":
                if attribute_config is None:
                    data = read_encoded_atr1(reader, size)
                else:
                    file.encoded_attributes = False
                    data = read_decoded_atr1(reader, attribute_config)

                attributes, size_per_attribute, string_table = data
                file.size_per_attribute = size_per_attribute
                file.attr_string_table = string_table
            case "TXT2":
                messages = read_txt2(reader, tag_config)
            case "TSY1":
                style_indexes = read_tsy1(reader, len(labels))
            case _:
                file.unsupported_sections[magic] = reader.read_bytes(size)

        file.section_list.append(magic)

    for i, label in labels.items():
        text = messages[i]
        attribute = None if attributes is None else attributes[i]
        style_index = None if style_indexes is None else style_indexes[i]
        file.entries.append(MSBTEntry(label, text, attribute, style_index))

    return file


def write_msbt(stream: BinaryIO, file: MSBT) -> None:
    """Writes a MSBT file to a stream.

    :param stream: the filestream.
    :param file: the MSBT file object.

    ## Usage
    ```
        with open(file_path, "wb") as file:
            write_msbt(file, msbt)
            ...
    ```
    """
    if stream is None:
        raise LMS_Exceptions.LMS_Error("Stream is not valid!")

    if not isinstance(file, MSBT):
        raise LMS_Exceptions.LMS_Error("File provided is not a MSBT.")

    writer = FileWriter(file.info.encoding)
    write_file_info(writer, "MsgStdBn", file.info)

    for section in file.section_list:
        match section:
            case "LBL1":
                labels = [entry.name for entry in file.entries]
                write_section(writer, "LBL1", write_labels, labels, file.slot_count)
            case "ATR1":
                attributes = [entry.attribute for entry in file.entries]
                if file.encoded_attributes:
                    write_section(
                        writer,
                        "ATR1",
                        write_encoded_atr1,
                        attributes,
                        file.size_per_attribute,
                        file.attr_string_table,
                    )
                else:
                    write_section(
                        writer,
                        "ATR1",
                        write_decoded_atr1,
                        attributes,
                        file.size_per_attribute,
                    )
            case "TXT2":
                messages = [entry.message for entry in file.entries]
                write_section(writer, "TXT2", write_txt2, messages)
            case "TSY1":
                style_indexes = [entry.style_index for entry in file.entries]
                write_section(writer, "TSY1", write_tsy1, style_indexes)
            case _:
                write_unsupported_section(
                    writer, section, file.unsupported_sections[section]
                )

    writer.seek(0x12)
    writer.write_uint32(writer.get_stream_size())
    stream.write(writer.get_data())



from enum import Enum
from typing import Literal


class FileEncoding(Enum):
    """An enum that represents a file encoding."""

    UTF8 = 0
    UTF16 = 1
    UTF32 = 2

    def to_string_format(
        self, big_endian: bool = False
    ):
        match self:
            case FileEncoding.UTF8:
                return "UTF-8"
            case FileEncoding.UTF16:
                if big_endian:
                    return "UTF-16-BE"
                return "UTF-16-LE"
            case FileEncoding.UTF32:
                if big_endian:
                    return "UTF-32-BE"
                return "UTF-32-LE"

    @property
    def width(self):
        match self:
            case FileEncoding.UTF8:
                return 1
            case FileEncoding.UTF16:
                return 2
            case FileEncoding.UTF32:
                return 4

    @property
    def terminator(self):
        return b"\x00" * self.width
